@@ -19,14 +19,18 @@
 
 #include <assert.h>
 #include <sys/select.h>
+#include <QCoreApplication>
+#include <QSocketNotifier>
+#include <QTimer>
 #include <QPointer>
 #include <QUrl>
 #include <curl/curl.h>
-#include "qjdnsshared.h"
 #include "bufferlist.h"
 #include "log.h"
+#include "addressresolver.h"
 
 #define BUFFER_SIZE 200000
+#define REQUEST_BODY_BUFFER_MAX 1000000
 
 // workaround for earlier curl versions
 #define UNPAUSE_WORKAROUND 1
@@ -61,13 +65,16 @@ public:
 	CURLSH *share;
 	CURL *easy;
 	QString method;
+	int maxRedirects;
 	bool expectBody;
+	bool alwaysSetBody;
 	bool bodyReadFrom;
 	struct curl_slist *dnsCache;
 	struct curl_slist *headersList;
 	int pauseBits;
 	BufferList in;
 	BufferList out;
+	int outPos;
 	bool inFinished;
 	bool outFinished;
 	bool haveStatusLine;
@@ -81,11 +88,14 @@ public:
 	CURLcode result;
 
 	CurlConnection() :
+		maxRedirects(-1),
 		expectBody(false),
+		alwaysSetBody(false),
 		bodyReadFrom(false),
 		dnsCache(NULL),
 		headersList(NULL),
 		pauseBits(0),
+		outPos(0),
 		inFinished(false),
 		outFinished(false),
 		haveStatusLine(false),
@@ -109,15 +119,21 @@ public:
 		curl_easy_setopt(easy, CURLOPT_WRITEDATA, this);
 		curl_easy_setopt(easy, CURLOPT_READFUNCTION, readFunction_cb);
 		curl_easy_setopt(easy, CURLOPT_READDATA, this);
+		curl_easy_setopt(easy, CURLOPT_SEEKFUNCTION, seekFunction_cb);
+		curl_easy_setopt(easy, CURLOPT_SEEKDATA, this);
 		curl_easy_setopt(easy, CURLOPT_HEADERFUNCTION, headerFunction_cb);
 		curl_easy_setopt(easy, CURLOPT_HEADERDATA, this);
 
-		curl_easy_setopt(easy, CURLOPT_BUFFERSIZE, BUFFER_SIZE);
+		curl_easy_setopt(easy, CURLOPT_BUFFERSIZE, (long)BUFFER_SIZE);
 		curl_easy_setopt(easy, CURLOPT_ACCEPT_ENCODING, "");
-		curl_easy_setopt(easy, CURLOPT_HTTP_CONTENT_DECODING, 1);
+		curl_easy_setopt(easy, CURLOPT_HTTP_CONTENT_DECODING, 1L);
 
 		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
-			curl_easy_setopt(easy, CURLOPT_VERBOSE, 1);
+			curl_easy_setopt(easy, CURLOPT_VERBOSE, 1L);
+
+#if LIBCURL_VERSION_NUM >= 0x072a00
+		curl_easy_setopt(easy, CURLOPT_PATH_AS_IS, 1L);
+#endif
 	}
 
 	~CurlConnection()
@@ -128,54 +144,64 @@ public:
 		curl_share_cleanup(share);
 	}
 
-	void setupMethod(const QString &_method)
+	void setupMethod(const QString &_method, bool _expectBody)
 	{
 		method = _method;
+		expectBody = _expectBody;
+		alwaysSetBody = false;
 
 		if(method == "OPTIONS")
 		{
-			expectBody = false;
+			assert(!expectBody);
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "OPTIONS");
 		}
 		else if(method == "HEAD")
 		{
-			expectBody = false;
+			assert(!expectBody);
+			curl_easy_setopt(easy, CURLOPT_NOBODY, 1L);
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, NULL);
 		}
 		else if(method == "GET")
 		{
-			expectBody = false;
-			curl_easy_setopt(easy, CURLOPT_HTTPGET, 1);
-			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, NULL);
+			if(!expectBody)
+			{
+				curl_easy_setopt(easy, CURLOPT_HTTPGET, 1L);
+				curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, NULL);
+			}
+			else
+			{
+				curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "GET");
+			}
 		}
 		else if(method == "POST")
 		{
-			expectBody = true;
-			//curl_easy_setopt(easy, CURLOPT_POST, 1);
+			alwaysSetBody = true;
+			//curl_easy_setopt(easy, CURLOPT_POST, 1L);
 			//curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, NULL);
-			curl_easy_setopt(easy, CURLOPT_UPLOAD, 1);
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "POST");
 		}
 		else if(method == "PUT")
 		{
-			expectBody = true;
-			curl_easy_setopt(easy, CURLOPT_UPLOAD, 1);
+			alwaysSetBody = true;
+			// PUT is implied by the UPLOAD option, which we set
+			//   below
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, NULL);
 		}
 		else if(method == "DELETE")
 		{
-			expectBody = false;
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "DELETE");
 		}
 		else
 		{
-			expectBody = true;
-			curl_easy_setopt(easy, CURLOPT_UPLOAD, 1);
+			alwaysSetBody = true;
 			curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, method.toLatin1().data());
 		}
+
+		if(expectBody || alwaysSetBody)
+			curl_easy_setopt(easy, CURLOPT_UPLOAD, 1L);
 	}
 
-	void setup(const QUrl &uri, const HttpHeaders &_headers, const QHostAddress &connectAddr = QHostAddress(), int connectPort = -1)
+	void setup(const QUrl &uri, const HttpHeaders &_headers, const QHostAddress &connectAddr = QHostAddress(), int connectPort = -1, int _maxRedirects = -1)
 	{
 		assert(!method.isEmpty());
 
@@ -214,8 +240,13 @@ public:
 			// curl will set this for us
 			headers.removeAll("Content-Length");
 		}
-		else if(expectBody)
-			chunked = true;
+		else
+		{
+			if(expectBody)
+				chunked = true;
+			else if(alwaysSetBody)
+				curl_easy_setopt(easy, CURLOPT_INFILESIZE_LARGE, (curl_off_t)0);
+		}
 
 		curl_slist_free_all(headersList);
 		foreach(const HttpHeader &h, headers)
@@ -223,11 +254,23 @@ public:
 			QByteArray i = h.first + ": " + h.second;
 			headersList = curl_slist_append(headersList, i.data());
 		}
-		if(chunked && !headers.contains("Transfer-Encoding"))
+
+		headers.removeAll("Transfer-Encoding");
+		if(chunked)
 			headersList = curl_slist_append(headersList, "Transfer-Encoding: chunked");
+
 		// disable expect usage as it appears to be buggy
 		curl_slist_append(headersList, "Expect:");
 		curl_easy_setopt(easy, CURLOPT_HTTPHEADER, headersList);
+
+		maxRedirects = _maxRedirects;
+		if(maxRedirects >= 0)
+		{
+			curl_easy_setopt(easy, CURLOPT_FOLLOWLOCATION, 1L);
+			curl_easy_setopt(easy, CURLOPT_MAXREDIRS, (long)maxRedirects);
+		}
+
+		curl_easy_setopt(easy, CURLOPT_POSTREDIR, CURL_REDIR_POST_ALL);
 	}
 
 	void update()
@@ -258,6 +301,12 @@ public:
 		return self->readFunction((char *)ptr, size * nmemb);
 	}
 
+	static int seekFunction_cb(void *userdata, curl_off_t offset, int origin)
+	{
+		CurlConnection *self = (CurlConnection *)userdata;
+		return self->seekFunction(offset, origin);
+	}
+
 	static size_t headerFunction_cb(void *ptr, size_t size, size_t nmemb, void *userdata)
 	{
 		CurlConnection *self = (CurlConnection *)userdata;
@@ -304,7 +353,27 @@ public:
 
 	size_t readFunction(char *p, size_t size)
 	{
-		QByteArray buf = out.take(size);
+		QByteArray buf;
+
+		if(outPos >= 0 && out.size() > REQUEST_BODY_BUFFER_MAX)
+		{
+			// exceeded buffer max, switch to unbuffered
+			QByteArray remaining = out.mid(outPos);
+			out.clear();
+			out += remaining;
+			outPos = -1;
+		}
+
+		if(outPos >= 0)
+		{
+			buf = out.mid(outPos, size);
+			outPos += buf.size();
+		}
+		else
+		{
+			buf = out.take(size);
+		}
+
 		if(!buf.isEmpty())
 		{
 			bodyReadFrom = true;
@@ -330,6 +399,35 @@ public:
 		}
 	}
 
+	int seekFunction(curl_off_t offset, int origin)
+	{
+		if(outPos < 0)
+		{
+			log_debug("seekFunction: can't seek. input is unbuffered");
+			return 1;
+		}
+
+		if(origin == SEEK_SET)
+		{
+			if(offset <= out.size())
+			{
+				outPos = offset;
+				log_debug("seekFunction: seeking to position %ld", offset);
+				return 0;
+			}
+			else
+			{
+				log_debug("seekFunction: %ld out of range (range: 0-%d)", offset, out.size());
+				return 1;
+			}
+		}
+		else
+		{
+			log_debug("seekFunction: unknown origin value: %d", origin);
+			return 1;
+		}
+	}
+
 	size_t headerFunction(char *p, size_t size)
 	{
 		assert(p[size - 1] == '\n');
@@ -342,7 +440,48 @@ public:
 			len = size - 1;
 
 		QByteArray line(p, len);
-		if(line.isEmpty())
+		if(!line.isEmpty())
+		{
+			if(haveResponseHeaders)
+			{
+				// does it look like we're getting a status
+				//   line again? (happens when redirecting)
+				int at = line.indexOf(' ');
+				if(at != -1 && !line.mid(0, at).contains(':'))
+				{
+					haveStatusLine = false;
+					haveResponseHeaders = false;
+					responseHeaders.clear();
+				}
+			}
+
+			if(!haveResponseHeaders)
+			{
+				if(haveStatusLine)
+				{
+					int at = line.indexOf(": ");
+					if(at == -1)
+						return -1;
+
+					log_debug("response header: %s", line.data());
+					responseHeaders += HttpHeader(line.mid(0, at), line.mid(at + 2));
+				}
+				else
+				{
+					// status reason we have to parse ourselves
+					int at = line.indexOf(' ');
+					if(at == -1)
+						return -1;
+					at = line.indexOf(' ', at + 1);
+					if(at == -1)
+						return -1;
+					responseReason = line.mid(at + 1);
+
+					haveStatusLine = true;
+				}
+			}
+		}
+		else
 		{
 			haveResponseHeaders = true;
 
@@ -356,42 +495,27 @@ public:
 				log_debug("got code 100, ignoring this header block");
 				haveStatusLine = false;
 				haveResponseHeaders = false;
+				responseHeaders.clear();
+				return size;
 			}
-			else
-			{
-				// if a content-encoding was used, don't provide content-length
-				QByteArray contentEncoding = responseHeaders.get("Content-Encoding");
-				if(!contentEncoding.isEmpty() && contentEncoding != "identity")
-					responseHeaders.removeAll("Content-Length");
 
-				// tell the app we've got the header block
-				update();
-			}
-		}
-		else if(!haveResponseHeaders)
-		{
-			if(haveStatusLine)
+			if(maxRedirects >= 0 && responseCode >= 300 && responseCode < 400 && responseHeaders.contains("Location"))
 			{
-				int at = line.indexOf(": ");
-				if(at == -1)
-					return -1;
-
-				log_debug("response header: %s", line.data());
-				responseHeaders += HttpHeader(line.mid(0, at), line.mid(at + 2));
+				log_debug("got code 3xx and redirects enabled, ignoring this header block");
+				haveStatusLine = false;
+				haveResponseHeaders = false;
+				responseHeaders.clear();
+				return size;
 			}
-			else
-			{
-				// status reason we have to parse ourselves
-				int at = line.indexOf(' ');
-				if(at == -1)
-					return -1;
-				at = line.indexOf(' ', at + 1);
-				if(at == -1)
-					return -1;
-				responseReason = line.mid(at + 1);
 
-				haveStatusLine = true;
-			}
+			// if a content-encoding was used, don't provide content-length
+			QByteArray contentEncoding = responseHeaders.get("Content-Encoding");
+			if(!contentEncoding.isEmpty() && contentEncoding != "identity")
+				responseHeaders.removeAll("Content-Length");
+
+			// tell the app we've got the header block
+			newlyReadOrEof = true;
+			update();
 		}
 
 		return size;
@@ -648,12 +772,16 @@ class HttpRequest::Private : public QObject
 public:
 	HttpRequest *q;
 	QJDnsShared *dns;
+	AddressResolver *resolver;
 	QString connectHost;
 	bool ignoreTlsErrors;
+	int maxRedirects;
 	HttpRequest::ErrorCondition errorCondition;
 	QString method;
 	QUrl uri;
 	HttpHeaders headers;
+	bool willWriteBody;
+	bool bodyNotAllowed;
 	QString host;
 	QList<QHostAddress> addrs;
 	HttpRequest::ErrorCondition mostSignificantError;
@@ -666,7 +794,10 @@ public:
 		q(_q),
 		dns(_dns),
 		ignoreTlsErrors(false),
+		maxRedirects(-1),
 		errorCondition(HttpRequest::ErrorNone),
+		willWriteBody(false),
+		bodyNotAllowed(false),
 		mostSignificantError(HttpRequest::ErrorGeneric),
 		ignoreBody(false),
 		conn(0),
@@ -674,6 +805,10 @@ public:
 	{
 		if(!g_man)
 			g_man = new CurlConnectionManager(QCoreApplication::instance());
+
+		resolver = new AddressResolver(dns, this);
+		connect(resolver, SIGNAL(resultsReady(const QList<QHostAddress> &)), SLOT(resolver_resultsReady(const QList<QHostAddress> &)));
+		connect(resolver, SIGNAL(error()), SLOT(resolver_error()));
 	}
 
 	~Private()
@@ -708,13 +843,13 @@ public:
 			conn = new CurlConnection;
 			connect(conn, SIGNAL(updated()), SLOT(conn_updated()));
 
-			conn->setupMethod(method);
+			conn->setupMethod(method, willWriteBody);
 			conn->out = out;
 			conn->outFinished = outFinished;
 		}
 	}
 
-	void start(const QString &_method, const QUrl &_uri, const HttpHeaders &_headers)
+	void start(const QString &_method, const QUrl &_uri, const HttpHeaders &_headers, bool _willWriteBody)
 	{
 		if(_method.isEmpty() || (_uri.scheme() != "https" && _uri.scheme() != "http"))
 		{
@@ -724,77 +859,86 @@ public:
 			return;
 		}
 
-		conn = new CurlConnection;
-		connect(conn, SIGNAL(updated()), SLOT(conn_updated()));
-
 		method = _method;
 		uri = _uri;
 		headers = _headers;
+		willWriteBody = _willWriteBody;
 
-		// eat any transport headers as they'd likely break things
-		headers.removeAll("Connection");
-		headers.removeAll("Keep-Alive");
-		headers.removeAll("Accept-Encoding");
-		headers.removeAll("Content-Encoding");
-		headers.removeAll("Transfer-Encoding");
-		headers.removeAll("Expect");
+		// we'd prefer not to send chunked encoding header if we don't
+		//   have to for certain methods, so wait and see if the user
+		//   actually tries to send a body before we start the request
+		if(willWriteBody && (method == "GET" || method == "DELETE"))
+			return;
 
-		conn->setupMethod(method);
+		// if the user might provide a body but we don't expect one
+		//   for the method type, then we'll wait and see what is
+		//   provided before starting the request.
+		if(willWriteBody && (method == "OPTIONS" || method == "HEAD"))
+		{
+			bodyNotAllowed = true;
+			return;
+		}
 
-		// if we aren't expecting a body, then we'll eat any input and start
-		//   the connect after endBody() is called.
-		if(conn->expectBody)
-			startConnect();
+		startConnect();
 	}
 
 	void writeBody(const QByteArray &body)
 	{
+		assert(willWriteBody);
+
 		if(body.isEmpty() || ignoreBody)
 			return;
 
-		assert(conn);
-
-		if(conn->expectBody)
-		{
-			conn->out += body;
-
-			if(conn->pauseBits & CURLPAUSE_SEND)
-			{
-				log_debug("send unpausing");
-				conn->pauseBits &= ~CURLPAUSE_SEND;
-				curl_easy_pause(conn->easy, conn->pauseBits);
-				g_man->update();
-			}
-		}
-		else
+		if(bodyNotAllowed)
 		{
 			ignoreBody = true;
 			errorCondition = HttpRequest::ErrorBodyNotAllowed;
 			QMetaObject::invokeMethod(q, "error", Qt::QueuedConnection);
+			return;
+		}
+
+		if(!conn)
+			startConnect();
+
+		assert(conn);
+
+		conn->out += body;
+
+		if(conn->pauseBits & CURLPAUSE_SEND)
+		{
+			log_debug("send unpausing");
+			conn->pauseBits &= ~CURLPAUSE_SEND;
+			curl_easy_pause(conn->easy, conn->pauseBits);
+			g_man->update();
 		}
 	}
 
 	void endBody()
 	{
+		assert(willWriteBody);
+
 		if(ignoreBody)
 			return;
 
+		if(!conn)
+		{
+			// if the user called endBody without actually writing
+			//   a body then we can set this flag off
+			willWriteBody = false;
+
+			startConnect();
+		}
+
 		assert(conn);
 
-		if(conn->expectBody)
+		conn->outFinished = true;
+
+		if(conn->pauseBits & CURLPAUSE_SEND)
 		{
-			if(conn->pauseBits & CURLPAUSE_SEND)
-			{
-				log_debug("send unpausing");
-				conn->outFinished = true;
-				conn->pauseBits &= ~CURLPAUSE_SEND;
-				curl_easy_pause(conn->easy, conn->pauseBits);
-				g_man->update();
-			}
-		}
-		else
-		{
-			startConnect();
+			log_debug("send unpausing");
+			conn->pauseBits &= ~CURLPAUSE_SEND;
+			curl_easy_pause(conn->easy, conn->pauseBits);
+			g_man->update();
 		}
 	}
 
@@ -822,23 +966,27 @@ public:
 
 	void startConnect()
 	{
+		assert(!conn);
+
+		conn = new CurlConnection;
+		connect(conn, SIGNAL(updated()), SLOT(conn_updated()));
+
+		// eat any transport headers as they'd likely break things
+		headers.removeAll("Connection");
+		headers.removeAll("Keep-Alive");
+		headers.removeAll("Accept-Encoding");
+		headers.removeAll("Content-Encoding");
+		headers.removeAll("Transfer-Encoding");
+		headers.removeAll("Expect");
+
+		conn->setupMethod(method, willWriteBody);
+
 		if(!connectHost.isEmpty())
 			host = connectHost;
 		else
 			host = uri.host();
 
-		QHostAddress addr(host);
-		if(!addr.isNull())
-		{
-			addrs += addr;
-			QMetaObject::invokeMethod(this, "tryNextAddress", Qt::QueuedConnection);
-		}
-		else
-		{
-			QJDnsSharedRequest *dreq = new QJDnsSharedRequest(dns);
-			connect(dreq, SIGNAL(resultsReady()), SLOT(dreq_resultsReady()));
-			dreq->query(QUrl::toAce(host), QJDns::A);
-		}
+		resolver->start(host);
 	}
 
 	// the idea with the priorities here is that an error is considered
@@ -880,12 +1028,12 @@ private slots:
 		if(!self)
 			return;
 
-		conn->setup(uri, headers, addr);
+		conn->setup(uri, headers, addr, -1, maxRedirects);
 
 		if(ignoreTlsErrors)
 		{
-			curl_easy_setopt(conn->easy, CURLOPT_SSL_VERIFYPEER, 0);
-			curl_easy_setopt(conn->easy, CURLOPT_SSL_VERIFYHOST, 0);
+			curl_easy_setopt(conn->easy, CURLOPT_SSL_VERIFYPEER, 0L);
+			curl_easy_setopt(conn->easy, CURLOPT_SSL_VERIFYHOST, 0L);
 		}
 
 		handleAdded = true;
@@ -895,28 +1043,16 @@ private slots:
 		g_man->doSocketAction(false, CURL_SOCKET_TIMEOUT, 0);
 	}
 
-	void dreq_resultsReady()
+	void resolver_resultsReady(const QList<QHostAddress> &results)
 	{
-		QJDnsSharedRequest *dreq = (QJDnsSharedRequest *)sender();
+		addrs += results;
+		tryNextAddress();
+	}
 
-		if(dreq->success())
-		{
-			QList<QJDns::Record> results = dreq->results();
-			foreach(const QJDns::Record &r, results)
-			{
-				if(r.type == QJDns::A)
-					addrs += r.address;
-			}
-
-			delete dreq;
-			tryNextAddress();
-		}
-		else
-		{
-			delete dreq;
-			errorCondition = HttpRequest::ErrorConnect;
-			emit q->error();
-		}
+	void resolver_error()
+	{
+		errorCondition = HttpRequest::ErrorConnect;
+		emit q->error();
 	}
 
 	void conn_updated()
@@ -941,6 +1077,9 @@ private slots:
 					//   was actually sent off
 					curError = HttpRequest::ErrorTimeout;
 					break;
+				case CURLE_TOO_MANY_REDIRECTS:
+					curError = HttpRequest::ErrorTooManyRedirects;
+					break;
 				default:
 					tryAgain = false;
 					curError = HttpRequest::ErrorGeneric;
@@ -1008,9 +1147,14 @@ void HttpRequest::setIgnoreTlsErrors(bool on)
 	d->ignoreTlsErrors = on;
 }
 
-void HttpRequest::start(const QString &method, const QUrl &uri, const HttpHeaders &headers)
+void HttpRequest::setFollowRedirects(int maxRedirects)
 {
-	d->start(method, uri, headers);
+	d->maxRedirects = maxRedirects;
+}
+
+void HttpRequest::start(const QString &method, const QUrl &uri, const HttpHeaders &headers, bool willWriteBody)
+{
+	d->start(method, uri, headers, willWriteBody);
 }
 
 void HttpRequest::writeBody(const QByteArray &body)
